#ifndef MINIMATH_H
#define MINIMATH_H

#include <math.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#define DEG_2_RAD (M_PI / 180.0)

/* Wrap arrays in structs, so we can allocate new objects in functions
 * and return them -- without the need for malloc(). */

/* A row-major 4x4 matrix. */
struct mat4 {
    GLfloat v[16];
};

/* A vector in R^3, components: 0 = x, 1 = y, 2 = z. */
struct vec3 {
    GLfloat v[3];
};

/* A quaternion, components: 0 = a, 1 = b*i, 2 = c*j, 3 = d*k.
 * For quaternion operations, see my PDF over here:
 * https://uninformativ.de/bin/SpaceSim-fc38848-CC-BY.pdf */
struct quat {
    GLfloat v[4];
};

struct mat4
mat4_identity(void)
{
    struct mat4 m = {0};

    m.v[0] = 1;
    m.v[5] = 1;
    m.v[10] = 1;
    m.v[15] = 1;

    return m;
}

struct mat4
mat4_perspective(GLfloat fov_deg, GLfloat aspect_w_over_h, GLfloat n, GLfloat f)
{
    struct mat4 m = {0};
    GLfloat fov_rad = fov_deg * DEG_2_RAD;

    /* n and f are meant to be *distances*, so they have to be positive
     * values. */

    /* This is a standard perspective projection matrix, as described by
     * version 2 or later of the PDF at [0], for example. However, as
     * most projections in the OpenGL world do, it includes an implicit
     * scaling of -1 along the Z axis. This flips the handedness. See
     * glFrustum(), for example, which does the same. It allows the user
     * to place objects in a right-handed coordinate system, while
     * OpenGL internally uses a left-handed one.
     *
     * [0]: https://uninformativ.de/git/asciifield */

    m.v[0] = 1.0 / (tan(0.5 * fov_rad) * aspect_w_over_h);
    m.v[5] = 1.0 / tan(0.5 * fov_rad);
    m.v[10] = -(f + n) / (f - n);
    m.v[11] = -(2.0 * f * n) / (f - n);
    m.v[14] = -1;

    return m;
}

struct mat4
mat4_rot_x(GLfloat angle_deg)
{
    struct mat4 m = {0};
    GLfloat angle_rad = angle_deg * DEG_2_RAD;

    m.v[0] = 1;

    m.v[5] = cos(angle_rad);
    m.v[6] = -sin(angle_rad);

    m.v[9] = sin(angle_rad);
    m.v[10] = cos(angle_rad);

    m.v[15] = 1;

    return m;
}

struct mat4
mat4_rot_y(GLfloat angle_deg)
{
    struct mat4 m = {0};
    GLfloat angle_rad = angle_deg * DEG_2_RAD;

    m.v[0] = cos(angle_rad);
    m.v[2] = sin(angle_rad);

    m.v[5] = 1;

    m.v[8] = -sin(angle_rad);
    m.v[10] = cos(angle_rad);

    m.v[15] = 1;

    return m;
}

struct mat4
mat4_rot_z(GLfloat angle_deg)
{
    struct mat4 m = {0};
    GLfloat angle_rad = angle_deg * DEG_2_RAD;

    m.v[0] = cos(angle_rad);
    m.v[1] = -sin(angle_rad);

    m.v[4] = sin(angle_rad);
    m.v[5] = cos(angle_rad);

    m.v[10] = 1;
    m.v[15] = 1;

    return m;
}

struct mat4
mat4_trans(GLfloat x, GLfloat y, GLfloat z)
{
    struct mat4 m = mat4_identity();

    m.v[3] = x;
    m.v[7] = y;
    m.v[11] = z;

    return m;
}

struct mat4
mat4_mult(struct mat4 A, struct mat4 B)
{
    struct mat4 m;

    m.v[0] = A.v[0]*B.v[0] + A.v[1]*B.v[4] + A.v[2]*B.v[8] + A.v[3]*B.v[12];
    m.v[4] = A.v[4]*B.v[0] + A.v[5]*B.v[4] + A.v[6]*B.v[8] + A.v[7]*B.v[12];
    m.v[8] = A.v[8]*B.v[0] + A.v[9]*B.v[4] + A.v[10]*B.v[8] + A.v[11]*B.v[12];
    m.v[12] = A.v[12]*B.v[0] + A.v[13]*B.v[4] + A.v[14]*B.v[8] + A.v[15]*B.v[12];

    m.v[1] = A.v[0]*B.v[1] + A.v[1]*B.v[5] + A.v[2]*B.v[9] + A.v[3]*B.v[13];
    m.v[5] = A.v[4]*B.v[1] + A.v[5]*B.v[5] + A.v[6]*B.v[9] + A.v[7]*B.v[13];
    m.v[9] = A.v[8]*B.v[1] + A.v[9]*B.v[5] + A.v[10]*B.v[9] + A.v[11]*B.v[13];
    m.v[13] = A.v[12]*B.v[1] + A.v[13]*B.v[5] + A.v[14]*B.v[9] + A.v[15]*B.v[13];

    m.v[2] = A.v[0]*B.v[2] + A.v[1]*B.v[6] + A.v[2]*B.v[10] + A.v[3]*B.v[14];
    m.v[6] = A.v[4]*B.v[2] + A.v[5]*B.v[6] + A.v[6]*B.v[10] + A.v[7]*B.v[14];
    m.v[10] = A.v[8]*B.v[2] + A.v[9]*B.v[6] + A.v[10]*B.v[10] + A.v[11]*B.v[14];
    m.v[14] = A.v[12]*B.v[2] + A.v[13]*B.v[6] + A.v[14]*B.v[10] + A.v[15]*B.v[14];

    m.v[3] = A.v[0]*B.v[3] + A.v[1]*B.v[7] + A.v[2]*B.v[11] + A.v[3]*B.v[15];
    m.v[7] = A.v[4]*B.v[3] + A.v[5]*B.v[7] + A.v[6]*B.v[11] + A.v[7]*B.v[15];
    m.v[11] = A.v[8]*B.v[3] + A.v[9]*B.v[7] + A.v[10]*B.v[11] + A.v[11]*B.v[15];
    m.v[15] = A.v[12]*B.v[3] + A.v[13]*B.v[7] + A.v[14]*B.v[11] + A.v[15]*B.v[15];

    return m;
}

struct vec3
mat4_mult_vec3(struct mat4 A, struct vec3 b)
{
    struct vec3 v;

    /* A corner case, where we pretend the matrix is a 3x3 matrix, so we
     * can multiply a vec3 with it. We simply ignore the matrix's last
     * row and column. */

    v.v[0] = A.v[0]*b.v[0] + A.v[1]*b.v[1] + A.v[2]*b.v[2];
    v.v[1] = A.v[4]*b.v[0] + A.v[5]*b.v[1] + A.v[6]*b.v[2];
    v.v[2] = A.v[8]*b.v[0] + A.v[9]*b.v[1] + A.v[10]*b.v[2];

    return v;
}

struct mat4
mat4_transposed(struct mat4 A)
{
    struct mat4 m;
    size_t r, c;

    for (r = 0; r < 4; r++)
        for (c = 0; c < 4; c++)
            m.v[r * 4 + c] = A.v[c * 4 + r];

    return m;
}

void
mat4_print(struct mat4 m)
{
    int i;

    for (i = 0; i < 16; i++)
    {
        if (i % 4 == 0)
            printf("\n");
        printf("%.2f ", m.v[i]);
    }
    printf("\n");
}

struct mat4
quat_to_mat4_rot(struct quat q)
{
    struct mat4 m;

    m.v[0] = 1 - 2 * (q.v[2]*q.v[2] + q.v[3]*q.v[3]);
    m.v[1] = 2 * (q.v[1]*q.v[2] - q.v[0]*q.v[3]);
    m.v[2] = 2 * (q.v[1]*q.v[3] + q.v[0]*q.v[2]);
    m.v[3] = 0;

    m.v[4] = 2 * (q.v[1]*q.v[2] + q.v[0]*q.v[3]);
    m.v[5] = 1 - 2 * (q.v[3]*q.v[3] + q.v[1]*q.v[1]);
    m.v[6] = 2 * (q.v[2]*q.v[3] - q.v[0]*q.v[1]);
    m.v[7] = 0;

    m.v[8] = 2 * (q.v[1]*q.v[3] - q.v[0]*q.v[2]);
    m.v[9] = 2 * (q.v[2]*q.v[3] + q.v[0]*q.v[1]);
    m.v[10] = 1 - 2 * (q.v[1]*q.v[1] + q.v[2]*q.v[2]);
    m.v[11] = 0;

    m.v[12] = 0;
    m.v[13] = 0;
    m.v[14] = 0;
    m.v[15] = 1;

    return m;
}

struct quat
quat_mult(struct quat A, struct quat B)
{
    struct quat q;

    q.v[0] = A.v[0]*B.v[0] - A.v[1]*B.v[1] - A.v[2]*B.v[2] - A.v[3]*B.v[3];
    q.v[1] = A.v[0]*B.v[1] + A.v[1]*B.v[0] + A.v[2]*B.v[3] - A.v[3]*B.v[2];
    q.v[2] = A.v[0]*B.v[2] - A.v[1]*B.v[3] + A.v[2]*B.v[0] + A.v[3]*B.v[1];
    q.v[3] = A.v[0]*B.v[3] + A.v[1]*B.v[2] - A.v[2]*B.v[1] + A.v[3]*B.v[0];

    return q;
}

struct quat
quat_for_rotation(GLfloat angle_deg, struct vec3 axis)
{
    GLfloat hPhi = angle_deg * DEG_2_RAD * 0.5;
    GLfloat sinHPhi = sin(hPhi);
    struct quat q;

    q.v[0] = cos(hPhi);
    q.v[1] = axis.v[0] * sinHPhi;
    q.v[2] = axis.v[1] * sinHPhi;
    q.v[3] = axis.v[2] * sinHPhi;

    return q;
}

struct quat
quat_normalized(struct quat q)
{
    GLfloat len;

    len = sqrt(q.v[0]*q.v[0] + q.v[1]*q.v[1] + q.v[2]*q.v[2] + q.v[3]*q.v[3]);
    q.v[0] /= len;
    q.v[1] /= len;
    q.v[2] /= len;
    q.v[3] /= len;

    return q;
}

GLfloat
vec3_length(struct vec3 v)
{
    return sqrt(v.v[0]*v.v[0] + v.v[1]*v.v[1] + v.v[2]*v.v[2]);
}

struct vec3
vec3_scaled(struct vec3 v, GLfloat s)
{
    v.v[0] *= s;
    v.v[1] *= s;
    v.v[2] *= s;

    return v;
}

struct vec3
vec3_add(struct vec3 a, struct vec3 b)
{
    a.v[0] += b.v[0];
    a.v[1] += b.v[1];
    a.v[2] += b.v[2];

    return a;
}

struct vec3
vec3_normalized(struct vec3 v)
{
    return vec3_scaled(v, 1.0 / vec3_length(v));
}

struct vec3
vec3_cross(struct vec3 a, struct vec3 b)
{
    struct vec3 v;

    v.v[0] = a.v[1]*b.v[2] - a.v[2]*b.v[1];
    v.v[1] = a.v[2]*b.v[0] - a.v[0]*b.v[2];
    v.v[2] = a.v[0]*b.v[1] - a.v[1]*b.v[0];

    return v;
}

#endif /* MINIMATH_H */
