/* Order matters, I've been told */
#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <arpa/inet.h>
#include <math.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "minimath.h"


struct FFImage
{
    uint32_t width, height;
    uint16_t *data;
} ff_colors, ff_height;

struct Crosshair
{
    GLuint vao, vbo;
} crosshair;

struct Globe
{
    GLuint vao, vbo, ebo;
    GLuint texture;
    GLfloat *vertices;
    GLuint *indices;
    size_t num_vertices, num_vertices_bytes;
    size_t num_indices, num_indices_bytes;

    int phi_verts, theta_verts;
    bool use_map;
    GLfloat map_r_range;
} globe = {
    .vertices = NULL,
    .indices = NULL,

    .phi_verts = 8,
    .theta_verts = 4,
    .use_map = false,
    .map_r_range = 0.1,
};

struct Shaders
{
    GLuint vertex, fragment, program;
    const GLchar *vertex_src, *fragment_src;
} shaders = {
    .vertex_src =
        "#version 330 core\n"
        "\n"
        "layout(location = 0) in vec3 in_position;"
        "layout(location = 1) in vec3 in_color;"
        "layout(location = 2) in vec3 in_texcoord;"
        "out vec3 vert_color;"
        "out vec3 vert_texcoord;"
        "uniform mat4 pvm;"
        "void main()"
        "{"
        "    gl_Position = pvm * vec4(in_position, 1.0f);"
        "    vert_color = in_color;"
        "    vert_texcoord = in_texcoord;"
        "}"
        ,
    .fragment_src =
        "#version 330 core\n"
        "\n"
        "in vec3 vert_color;"
        "in vec3 vert_texcoord;"
        "out vec4 frag_color;"
        "uniform sampler2D the_texture;"
        "void main()"
        "{"
        "    frag_color = texture(the_texture, vert_texcoord.xy) * vert_texcoord.z +"
        "                 vec4(vert_color, 1.0f) * (1 - vert_texcoord.z);"
        "}"
        ,
};

struct Window
{
    GLFWwindow* glfw_win;
    int w, h;
    bool mouse_look, mouse_look_initialized;
    double last_mouse_x, last_mouse_y;
    GLfloat mouse_speed;
    bool keys_down[GLFW_KEY_LAST], mouse_down[GLFW_MOUSE_BUTTON_LAST];
} win = {
    .w = 800,
    .h = 600,
    .mouse_look = false,
    .mouse_look_initialized = false,
    .mouse_speed = 0.05,
    .keys_down = {0},
    .mouse_down = {0},
};

struct Camera
{
    GLfloat friction_coeff, accel_coeff, friction_coeff_rot, accel_coeff_rot;
    struct vec3 speed;
    struct vec3 pos;
    struct quat ori;
    GLfloat fov;
    GLfloat speed_rot;
    struct mat4 mat_proj;
} cam = {
    .speed = {.v = {0, 0, 0}},
    .pos = {.v = {0, 0, 2}},     /* some units away from the globe */
    .ori = {.v = {1, 0, 0, 0}},  /* identity quaternion */
    .fov = 80,
    .speed_rot = 0,
    .accel_coeff_rot = 180,
};

enum VertexDataType
{
    VDT_VERTEX,
    VDT_COLOR,
    VDT_TEXCOORD,
};

const struct vec3 X_AXIS = {.v = {1, 0, 0}};
const struct vec3 Y_AXIS = {.v = {0, 1, 0}};
const struct vec3 Z_AXIS = {.v = {0, 0, 1}};


bool
ff_load(char *path, struct FFImage *image)
{
    FILE *fp = NULL;
    uint32_t hdr[4], x, y, channel;

    fp = fopen(path, "r");
    if (!fp)
    {
        fprintf(stderr, "Could not open image file '%s'\n", path);
        goto errout;
    }

    if (fread(hdr, sizeof (uint32_t), 4, fp) != 4)
    {
        fprintf(stderr, "Could not read farbfeld header from '%s'\n",
                path);
        goto errout;
    }

    if (memcmp("farbfeld", hdr, (sizeof "farbfeld") - 1) != 0)
    {
        fprintf(stderr, "Magic number is not 'farbfeld', path '%s'\n",
                path);
        goto errout;
    }

    image->width = ntohl(hdr[2]);
    image->height = ntohl(hdr[3]);

    image->data = calloc(image->width * image->height * 4, sizeof (uint16_t));
    if (!image->data)
    {
        fprintf(stderr, "Could not allocate memory, image->data for '%s'\n",
                path);
        goto errout;
    }

    if (fread(image->data,
              image->width * image->height * 4 * sizeof (uint16_t), 1, fp) != 1)
    {
        fprintf(stderr, "Unexpected EOF when reading '%s'\n", path);
        goto errout;
    }

    for (y = 0; y < image->height; y++)
        for (x = 0; x < image->width; x++)
            for (channel = 0; channel < 4; channel++)
                image->data[(y * image->width + x) * 4 + channel] =
                    ntohs(image->data[(y * image->width + x) * 4 + channel]);

    fclose(fp);
    return true;

errout:
    if (image)
        free(image->data);

    if (fp)
        fclose(fp);

    return false;
}

uint16_t
ff_get_clamped(struct FFImage *image, uint32_t x, uint32_t y, uint8_t channel)
{
    size_t idx;

    x = x >= image->width ? image->width - 1 : x;
    y = y >= image->height ? image->height - 1 : y;
    channel = channel > 3 ? 3 : channel;

    idx = (y * image->width + x) * 4 + channel;

    return image->data[idx];
}

void
globe_vertex_data(size_t vertex_index, enum VertexDataType kind,
                  GLfloat a, GLfloat b, GLfloat c)
{
    size_t offset;

    offset = vertex_index * 9;
    switch (kind)
    {
        case VDT_VERTEX:
            /* nop */
            break;
        case VDT_COLOR:
            offset += 3;
            break;
        case VDT_TEXCOORD:
            offset += 6;
            break;
    }

    globe.vertices[offset + 0] = a;
    globe.vertices[offset + 1] = b;
    globe.vertices[offset + 2] = c;
}

void
globe_vertex_generate(size_t vertex_index, int theta_i, int phi_i)
{
    GLfloat theta, phi;
    int phi_u_hei, theta_v_hei;
    GLfloat phi_tex, theta_tex;
    GLfloat r;

    /* theta_i \in [0, globe.theta_verts)  and
     * phi_i   \in [0, globe.phi_verts)
     *
     * theta must span the full 180 degree, so the last theta_i that we
     * see must be mapped to M_PI. The last one that we see is
     * globe.theta_verts - 1.
     *
     * Similarly, phi must cover the full 360 degree, so the last phi_i
     * that we see must be mapped to 2 * M_PI.
     *
     * This means that the will be one column of duplicate vertex
     * positions: At the left/right edge of the image map. One column
     * will be at 0 degree, the other at 360 degree (which is the same
     * position, of course). Those two columns will be assigned two
     * different sets of texture coordinates, though.
     *
     * This will result in a "broken" mesh at the seem due to different
     * heights. I see no way to avoid this when using texture mapping.
     * (See earlier versions of this file where we used vertex colors.
     * That allowed us to have a closed mesh.) */
    theta = (GLfloat)theta_i / (globe.theta_verts - 1) * M_PI;
    phi = (GLfloat)phi_i / (globe.phi_verts - 1) * 2 * M_PI;

    phi_u_hei = (GLfloat)phi_i / globe.phi_verts * ff_height.width;
    theta_v_hei = (GLfloat)theta_i / globe.theta_verts * ff_height.height;

    phi_tex = (GLfloat)phi_i / (globe.phi_verts - 1);
    theta_tex = (GLfloat)theta_i / (globe.theta_verts - 1);

    if (globe.use_map)
    {
        r = ff_get_clamped(&ff_height, phi_u_hei, theta_v_hei, 0) / 65535.0;
        r *= globe.map_r_range;
        r += 1;
    }
    else
    {
        globe_vertex_data(vertex_index, VDT_COLOR,
            (GLfloat)phi_i / globe.phi_verts,
            (GLfloat)theta_i / globe.theta_verts,
            0.3);
        r = 1;
    }

    /* Erect the globe right from the start. Don't bother rotating this
     * later on. This flips the sign of the z component and switches y
     * and z component. */
    globe_vertex_data(vertex_index, VDT_VERTEX,
        r * sin(theta) * cos(phi),
        r * cos(theta),
        -r * sin(theta) * sin(phi));

    /* Texture coordinate. Note that each vertex comprises 9 (!) float
     * values: 3 for position, 3 for color, and 3 for texture. Normally,
     * texture coordinates are 2D, so you'd only need 2 floats. We abuse
     * the third float to indicate whether the texture should be used,
     * or the color. If it's 1.0, the texture is used. 0.0 for the
     * color. The fragment shader handles this. */
    globe_vertex_data(vertex_index, VDT_TEXCOORD, phi_tex, theta_tex,
                      globe.use_map ? 1.0 : 0.0);
}

void
map_load_images(void)
{
    uint8_t *ff_colors_compat;
    uint32_t x, y, channel;

    fprintf(stderr, "Loading map images ...\n");

    if (!ff_load("world_colors.ff", &ff_colors))
        exit(EXIT_FAILURE);

    if (!ff_load("world_height.ff", &ff_height))
        exit(EXIT_FAILURE);

    fprintf(stderr, "Done.\n");

    fprintf(stderr, "Generating texture ...\n");
    ff_colors_compat = calloc(ff_colors.width * ff_colors.height, 4 * sizeof (uint8_t));
    if (ff_colors_compat == NULL)
    {
        perror("calloc ff_colors_compat");
        exit(EXIT_FAILURE);
    }
    for (y = 0; y < ff_colors.height; y++)
        for (x = 0; x < ff_colors.width; x++)
            for (channel = 0; channel < 4; channel++)
                ff_colors_compat[(y * ff_colors.width + x) * 4 + channel] =
                    ff_colors.data[(y * ff_colors.width + x) * 4 + channel] / 65535.0 * 255;

    /* Not sure if needed */
    /*glActiveTexture(GL_TEXTURE0);
    glUniform1i(glGetUniformLocation(shaders.program, "the_texture"), 0);*/

    glBindTexture(GL_TEXTURE_2D, globe.texture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ff_colors.width, ff_colors.height, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, ff_colors_compat);
    glGenerateMipmap(GL_TEXTURE_2D);
    free(ff_colors_compat);

    fprintf(stderr, "Done.\n");
}

void
map_build(void)
{
    int theta_i, phi_i;
    size_t idx;

    fprintf(stderr, "Building new map for: theta_verts %d, phi_verts %d\n",
            globe.theta_verts, globe.phi_verts);

    free(globe.vertices);
    free(globe.indices);

    /* XXX This will overflow at some point. */
    globe.num_vertices = globe.theta_verts * globe.phi_verts;
    globe.num_vertices_bytes = globe.num_vertices * (3 + 3 + 3) * sizeof (GLfloat);
    globe.vertices = calloc(globe.num_vertices_bytes, 1);
    if (globe.vertices == NULL)
    {
        perror("calloc globe.vertices");
        exit(EXIT_FAILURE);
    }

    /* XXX This will overflow at some point. */
    globe.num_indices = (globe.theta_verts - 1) * (globe.phi_verts - 1) * 2 * 3;
    globe.num_indices_bytes = globe.num_indices * sizeof (GLuint);
    globe.indices = calloc(globe.num_indices_bytes, 1);
    if (globe.indices == NULL)
    {
        perror("calloc globe.indices");
        exit(EXIT_FAILURE);
    }

    fprintf(stderr, "Allocated: %zu vertices (%zu bytes) with %zu indices (%zu bytes)\n",
            globe.num_vertices, globe.num_vertices_bytes,
            globe.num_indices, globe.num_indices_bytes);
    fprintf(stderr, "Triangles: %zu\n", globe.num_indices / 3);

    idx = 0;
    for (theta_i = 0; theta_i < globe.theta_verts; theta_i++)
        for (phi_i = 0; phi_i < globe.phi_verts; phi_i++)
            globe_vertex_generate(idx++, theta_i, phi_i);

    idx = 0;
    for (theta_i = 0; theta_i < globe.theta_verts - 1; theta_i++)
    {
        /* Do not connect the first and the last column. As explained in
         * globe_vertex_generate(), they share the same value of phi,
         * but have different texture coordinates. We could connect
         * them, but there would be a texture glitch. If we do not
         * connect them, the mesh will not be closed (due to different
         * heights). Either way, it sucks a little bit. */
        for (phi_i = 0; phi_i < globe.phi_verts - 1; phi_i++)
        {
            globe.indices[idx++] =  theta_i      * globe.phi_verts +  phi_i;
            globe.indices[idx++] = (theta_i + 1) * globe.phi_verts +  phi_i;
            globe.indices[idx++] = (theta_i + 1) * globe.phi_verts + (phi_i + 1);

            globe.indices[idx++] =  theta_i      * globe.phi_verts +  phi_i;
            globe.indices[idx++] = (theta_i + 1) * globe.phi_verts + (phi_i + 1);
            globe.indices[idx++] =  theta_i      * globe.phi_verts + (phi_i + 1);
        }
    }

    fprintf(stderr, "Map constructed.\n");

    /* Bind everything to our VAO. In display(), we can then bind this
     * VAO and have everything ready. */
    glBindVertexArray(globe.vao);

    /* We want to use this VBO from now on. */
    glBindBuffer(GL_ARRAY_BUFFER, globe.vbo);

    /* Transfer vertex and color data. */
    glBufferData(GL_ARRAY_BUFFER, globe.num_vertices_bytes, globe.vertices, GL_STATIC_DRAW);

    /* Bind and transfer index data. */
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, globe.ebo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, globe.num_indices_bytes, globe.indices, GL_STATIC_DRAW);

    /* The first "0" here defines the "attribute index". Look at the
     * GLSL code for the vertex shader above: It uses "location = 0".
     * That's the same "0".
     *
     * We have to enable it, so it can actually be used in the shader.
     * We then tell OpenGL about the format of the data. Both of these
     * operate implicitly on the currently bound GL_ARRAY_BUFFER.
     *
     * Since there are also color values in globe, the vertex data is no
     * longer tightly packed and we have to set "stride" and "pointer"
     * accordingly. */
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE,
                          9 * sizeof (GLfloat), 0);
    /* And here come our colors. */
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE,
                          9 * sizeof (GLfloat), (void *)(3 * sizeof (GLfloat)));
    /* Finally texture coordinates. */
    glEnableVertexAttribArray(2);
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE,
                          9 * sizeof (GLfloat), (void *)(6 * sizeof (GLfloat)));
}

void
display(void)
{
    GLuint pvm_loc;
    struct mat4 pvm_mat;

    pvm_mat = cam.mat_proj;
    pvm_mat = mat4_mult(pvm_mat, quat_to_mat4_rot(cam.ori));
    pvm_mat = mat4_mult(pvm_mat, mat4_trans(-cam.pos.v[0], -cam.pos.v[1], -cam.pos.v[2]));

    glClearColor(0, 0, 0.2, 1);
    glClear(GL_COLOR_BUFFER_BIT);

    glEnable(GL_DEPTH_TEST);
    glClear(GL_DEPTH_BUFFER_BIT);

    glFrontFace(GL_CCW);
    glEnable(GL_CULL_FACE);

    glUseProgram(shaders.program);

    /* pvm = Projection, View, Model. I think this is better than the
     * common "MVP" acronym, since you multiply vectors from the right.
     * So, when you read "pvm * V", "V" first goes through "M", then
     * through "V", and then through "P". */
    pvm_loc = glGetUniformLocation(shaders.program, "pvm");
    /* GL_TRUE to transpose the matrix. When we write pvm_mat as we did
     * above, it's row-major order. OpenGL expects column-major order by
     * default, so we need to transpose it. */
    glUniformMatrix4fv(pvm_loc, 1, GL_TRUE, pvm_mat.v);

    /* Bind the VAO for our globe with all the attributes that we
     * prepared in map_build(). Then draw it. */
    glBindTexture(GL_TEXTURE_2D, globe.texture);
    glBindVertexArray(globe.vao);
    glDrawElements(GL_TRIANGLES, globe.num_indices, GL_UNSIGNED_INT, NULL);

    /* Draw the crosshair. (Just account for current window size, we
     * want the crosshair to be exactly 20 pixels wide and high.) */
    glDisable(GL_DEPTH_TEST);
    pvm_mat = mat4_identity();
    pvm_mat.v[0] = 10.0 / (0.5 * win.w);
    pvm_mat.v[5] = 10.0 / (0.5 * win.h);
    glUniformMatrix4fv(pvm_loc, 1, GL_TRUE, pvm_mat.v);
    glBindVertexArray(crosshair.vao);
    glDrawArrays(GL_LINES, 0, 4);
}

void
cam_update_accel_and_friction_coeffs(void)
{
    GLfloat dist_from_surface;

    dist_from_surface = vec3_length(cam.pos);
    dist_from_surface -= 1;  /* globe radius */
    dist_from_surface = dist_from_surface < 0 ? 0 : dist_from_surface;

    cam.accel_coeff = dist_from_surface;
    cam.accel_coeff = cam.accel_coeff > 10 ? 10 : cam.accel_coeff;
    cam.accel_coeff = cam.accel_coeff < 0.15 ? 0.15 : cam.accel_coeff;

    cam.friction_coeff = 1.0 / dist_from_surface;
    cam.friction_coeff = cam.friction_coeff > 2.5 ? 2.5 : cam.friction_coeff;
    cam.friction_coeff = cam.friction_coeff < 0.1 ? 0.1 : cam.friction_coeff;

    cam.friction_coeff_rot = 1.0 / dist_from_surface;
    cam.friction_coeff_rot = cam.friction_coeff_rot > 4 ? 4 : cam.friction_coeff_rot;
    cam.friction_coeff_rot = cam.friction_coeff_rot < 1 ? 1 : cam.friction_coeff_rot;
}

void
cam_update(GLfloat dt)
{
    double dx, dy, xpos, ypos;
    struct vec3 accel_friction, accel_move = {0};
    GLfloat accel_friction_rot, accel_rot = 0, brake_coeff = 1;
    struct mat4 ori_g_to_l, ori_l_to_g;
    struct vec3 speed_world, x_axis_global, y_axis_global, z_axis_global;
    struct quat r;

    cam_update_accel_and_friction_coeffs();

    /* The camera has a position (units). It changes this position
     * according to some speed (units per second). That speed will be
     * changed according to some acceleration (units per second per
     * second).
     *
     * Let the camera's mass be 1. The forces that cause the
     * acceleration are not specified (except for friction), key presses
     * just cause some acceleration.
     *
     * Kinetic friction always tries to get things to a halt, so it
     * points in the opposite direction as the current speed.
     *
     * May be of interest:
     *
     * https://en.wikipedia.org/wiki/Acceleration#Uniform_acceleration
     * https://en.wikipedia.org/wiki/Friction */

    /* quat_to_mat4_rot() returns a matrix that transforms from global
     * coordinates to local camera coordinates. We want to do the
     * opposite now, since cam.speed is local camera coordinates. */
    ori_g_to_l = quat_to_mat4_rot(cam.ori);
    ori_l_to_g = mat4_transposed(ori_g_to_l);

    /* Keyboard -- movement */
    if (win.keys_down[GLFW_KEY_W])
        accel_move.v[2] = -cam.accel_coeff;
    else if (win.keys_down[GLFW_KEY_S])
        accel_move.v[2] = cam.accel_coeff;
    if (win.keys_down[GLFW_KEY_A])
        accel_move.v[0] = -cam.accel_coeff;
    else if (win.keys_down[GLFW_KEY_D])
        accel_move.v[0] = cam.accel_coeff;
    if (win.keys_down[GLFW_KEY_LEFT_SHIFT])
        accel_move.v[1] = cam.accel_coeff;
    else if (win.keys_down[GLFW_KEY_LEFT_CONTROL])
        accel_move.v[1] = -cam.accel_coeff;

    if (win.mouse_down[GLFW_MOUSE_BUTTON_LEFT])
        accel_move = vec3_scaled(accel_move, 5);

    if (win.mouse_down[GLFW_MOUSE_BUTTON_RIGHT])
        brake_coeff = 5;

    accel_friction = vec3_scaled(cam.speed, -cam.friction_coeff * brake_coeff);
    cam.speed = vec3_add(cam.speed, vec3_scaled(accel_move, dt));
    cam.speed = vec3_add(cam.speed, vec3_scaled(accel_friction, dt));

    speed_world = mat4_mult_vec3(ori_l_to_g, cam.speed);
    speed_world = vec3_scaled(speed_world, dt);
    cam.pos = vec3_add(cam.pos, speed_world);

    /* Keyboard -- roll */
    if (win.keys_down[GLFW_KEY_Q])
        accel_rot = -cam.accel_coeff_rot;
    if (win.keys_down[GLFW_KEY_E])
        accel_rot = cam.accel_coeff_rot;

    if (win.mouse_down[GLFW_MOUSE_BUTTON_LEFT])
        accel_rot *= 5;

    accel_friction_rot = cam.speed_rot * (-cam.friction_coeff_rot) * brake_coeff;
    cam.speed_rot += accel_rot * dt;
    cam.speed_rot += accel_friction_rot * dt;

    z_axis_global = mat4_mult_vec3(ori_l_to_g, Z_AXIS);
    r = quat_for_rotation(cam.speed_rot * dt, z_axis_global);
    cam.ori = quat_mult(cam.ori, r);

    /* Mouse -- orientation */
    if (win.mouse_look)
    {
        glfwGetCursorPos(win.glfw_win, &xpos, &ypos);
        dx = xpos - win.last_mouse_x;
        dy = ypos - win.last_mouse_y;
        win.last_mouse_x = xpos;
        win.last_mouse_y = ypos;

        if (!win.mouse_look_initialized)
        {
            /* First time we run into this hook since the user activated
             * mouse look, so win.last_mouse_* are meaningless and so
             * are dx and dy. */
            win.mouse_look_initialized = true;
        }
        else
        {
            y_axis_global = mat4_mult_vec3(ori_l_to_g, Y_AXIS);
            r = quat_for_rotation(dx * win.mouse_speed, y_axis_global);
            cam.ori = quat_mult(cam.ori, r);

            /* -dy because mouse look inverted. I can't stand anything
             * else. */
            x_axis_global = mat4_mult_vec3(ori_l_to_g, X_AXIS);
            r = quat_for_rotation(-dy * win.mouse_speed, x_axis_global);
            cam.ori = quat_mult(cam.ori, r);
        }
    }

    /* Now that we're done with all the updates to cam.ori, we can
     * normalize it. */
    cam.ori = quat_normalized(cam.ori);
}

void
keyboard(GLFWwindow *glfw_win, int key, int scancode, int action, int mods)
{
    bool needs_rebuild = false;

    (void)glfw_win;
    (void)scancode;
    (void)mods;

    /* Movement keys first. They don't do anything but set a flag, so we
     * can easily have multiple keys pressed at the same time and
     * process them all simultaneously. This will be done in
     * cam_update(). */

    switch (key)
    {
        case GLFW_KEY_W:
        case GLFW_KEY_S:
        case GLFW_KEY_A:
        case GLFW_KEY_D:
        case GLFW_KEY_Q:
        case GLFW_KEY_E:
        case GLFW_KEY_LEFT_SHIFT:
        case GLFW_KEY_LEFT_CONTROL:
        case GLFW_KEY_SPACE:
            if (action == GLFW_PRESS)
                win.keys_down[key] = true;
            else if (action == GLFW_RELEASE)
                win.keys_down[key] = false;
            return;
    }

    /* Simple keys follow. */

    if (action != GLFW_PRESS && action != GLFW_REPEAT)
        return;

    switch (key)
    {
        case GLFW_KEY_P:
            if (mods & GLFW_MOD_SHIFT)
                globe.phi_verts -= 2;
            else
                globe.phi_verts += 2;
            needs_rebuild = true;
            break;
        case GLFW_KEY_T:
            if (mods & GLFW_MOD_SHIFT)
                globe.theta_verts--;
            else
                globe.theta_verts++;
            needs_rebuild = true;
            break;
        case GLFW_KEY_I:
            if (mods & GLFW_MOD_SHIFT)
            {
                globe.theta_verts--;
                globe.phi_verts = globe.theta_verts * 2;
            }
            else
            {
                globe.theta_verts++;
                globe.phi_verts = globe.theta_verts * 2;
            }
            needs_rebuild = true;
            break;
        case GLFW_KEY_H:
            if (mods & GLFW_MOD_SHIFT)
            {
                globe.theta_verts = 4;
                globe.phi_verts = 8;
            }
            else
            {
                globe.theta_verts = HEIGHTMAP_HEIGHT;
                globe.phi_verts = HEIGHTMAP_WIDTH;
            }
            needs_rebuild = true;
            break;
        case GLFW_KEY_M:
            globe.use_map = !globe.use_map;
            needs_rebuild = true;
            break;
        case GLFW_KEY_R:
            if (mods & GLFW_MOD_SHIFT)
                globe.map_r_range -= 0.01;
            else
                globe.map_r_range += 0.01;
            needs_rebuild = true;
            break;
        case GLFW_KEY_L:
            win.mouse_look = !win.mouse_look;
            if (win.mouse_look)
            {
                win.mouse_look_initialized = false;
                glfwSetInputMode(win.glfw_win, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
            }
            else
                glfwSetInputMode(win.glfw_win, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
            break;
        case GLFW_KEY_ESCAPE:
            exit(EXIT_SUCCESS);
    }

    if (needs_rebuild)
        map_build();
}

void
mouse(GLFWwindow *glfw_win, int button, int action, int mods)
{
    (void)glfw_win;
    (void)mods;

    win.mouse_down[button] = (action == GLFW_PRESS);
}

void
reshape(GLFWwindow *glfw_win, int width, int height)
{
    (void)glfw_win;

    glViewport(0, 0, width, height);
    cam.mat_proj = mat4_perspective(cam.fov, (GLfloat)width / height, 0.01, 10.0);

    win.w = width;
    win.h = height;
}

void
setup_opengl_objects(void)
{
    /* https://www.khronos.org/opengl/wiki/Vertex_Specification#Vertex_Array_Object
     *
     * "A Vertex Array Object (VAO) is an OpenGL Object that stores all
     * of the state needed to supply vertex data (with one minor
     * exception noted below)."
     *
     * We only need to create and bind it once. It will then be used
     * throughout the rest of the program. */
    glGenVertexArrays(1, &globe.vao);

    /* Where we will store our vertices. */
    glGenBuffers(1, &globe.vbo);

    /* If we were to draw and store each triangle individually, we'd
     * waste a lot of space. We thus use an element buffer object and
     * shared vertices. */
    glGenBuffers(1, &globe.ebo);

    /* Now the same for our crosshair. */
    glGenVertexArrays(1, &crosshair.vao);
    glGenBuffers(1, &crosshair.vbo);

    /* Globe texture. */
    glGenTextures(1, &globe.texture);
}

void
print_shader_compilation_errors(GLuint which)
{
    int success;
    char info_log[512];

    glGetShaderiv(which, GL_COMPILE_STATUS, &success);
    if (success == 0)
    {
        glGetShaderInfoLog(which, sizeof info_log, NULL, info_log);
        info_log[sizeof info_log - 1] = 0;
        fprintf(stderr, "Compiling shader %u failed: '%s'\n", which, info_log);
    }
}

void
print_shader_linker_errors(GLuint which)
{
    int success;
    char info_log[512];

    glGetProgramiv(which, GL_LINK_STATUS, &success);
    if (success == 0)
    {
        glGetProgramInfoLog(which, sizeof info_log, NULL, info_log);
        info_log[sizeof info_log - 1] = 0;
        fprintf(stderr, "Linking program failed: '%s'\n", info_log);
    }
}

void
setup_shaders(void)
{
    shaders.vertex = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(shaders.vertex, 1, &shaders.vertex_src, NULL);
    glCompileShader(shaders.vertex);
    print_shader_compilation_errors(shaders.vertex);

    shaders.fragment = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(shaders.fragment, 1, &shaders.fragment_src, NULL);
    glCompileShader(shaders.fragment);
    print_shader_compilation_errors(shaders.fragment);

    shaders.program = glCreateProgram();
    glAttachShader(shaders.program, shaders.vertex);
    glAttachShader(shaders.program, shaders.fragment);
    glLinkProgram(shaders.program);
    print_shader_linker_errors(shaders.program);
}

void
gl_init(void)
{
    if (!glfwInit())
    {
        fprintf(stderr, "Failed to initialize GLFW\n");
        exit(EXIT_FAILURE);
    }

    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);

    /* Enable forward-compatible mode, which disables all "legacy"
     * functions that have been marked as deprecated since OpenGL 3.0. */
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

    /* Disable deprecated functions ... again? Not sure if we need both
     * calls. */
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    win.glfw_win = glfwCreateWindow(win.w, win.h, "Hello World", NULL, NULL);
    if (!win.glfw_win)
    {
        fprintf(stderr, "Failed to create glfw_win\n");
        glfwTerminate();
        exit(EXIT_FAILURE);
    }

    glfwMakeContextCurrent(win.glfw_win);

    if (glewInit() != GLEW_OK)
    {
        fprintf(stderr, "Failed to initialize GLEW\n");
        glfwTerminate();
        exit(EXIT_FAILURE);
    }

    reshape(win.glfw_win, win.w, win.h);
    glfwSetFramebufferSizeCallback(win.glfw_win, reshape);
    glfwSetKeyCallback(win.glfw_win, keyboard);
    glfwSetMouseButtonCallback(win.glfw_win, mouse);

    setup_opengl_objects();
    setup_shaders();
}

void
gl_main_loop(void)
{
    double dt, t_fps_last, t, t_last = glfwGetTime();
    int frames = 0;

    t_fps_last = t_last;

    while (!glfwWindowShouldClose(win.glfw_win))
    {
        display();
        glfwSwapBuffers(win.glfw_win);
        frames++;

        glfwPollEvents();

        t = glfwGetTime();
        dt = t - t_last;
        cam_update(dt);
        t_last = t;

        if (t - t_fps_last >= 1)
        {
            fprintf(stderr, "fps %d\n", frames);
            frames = 0;
            t_fps_last = t;
        }
    }

    glfwTerminate();
}

void
crosshair_build(void)
{
    /* pos = x, y, z
     * col = r, g, b
     * tex = s, t, active
     *
     * All values for "active" are 0, which means that no texture shall
     * be used. */
    GLfloat ch[] = {
        /* pos      col        tex */
        -1, 0, 0,   1, 0, 0,   0, 0, 0,
         1, 0, 0,   1, 0, 0,   0, 0, 0,

         0, -1, 0,   1, 0, 0,   0, 0, 0,
         0,  1, 0,   1, 0, 0,   0, 0, 0,
    };

    glBindVertexArray(crosshair.vao);
    glBindBuffer(GL_ARRAY_BUFFER, crosshair.vbo);
    glBufferData(GL_ARRAY_BUFFER, sizeof ch, ch, GL_STATIC_DRAW);

    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE,
                          9 * sizeof (GLfloat), 0);
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE,
                          9 * sizeof (GLfloat), (void *)(3 * sizeof (GLfloat)));
    glEnableVertexAttribArray(2);
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE,
                          9 * sizeof (GLfloat), (void *)(3 * sizeof (GLfloat)));
}

void
hello(void)
{
    puts("-----------");
    puts("HELLO WORLD");
    puts("-----------");
    puts("");
    puts("Important keys:");
    puts("");
    puts("-   m: Toggle usage of colors/heights from image maps");
    puts("-   l: Toggle mouse look");
    puts("-   h: Activate detailed mesh");
    puts("");
    puts("-   w, a, s, d: Move camera");
    puts("-   Shift, control: Move camera");
    puts("-   q, e: Roll camera");
    puts("-   Left mouse button: Speed booster");
    puts("-   Right mouse button: Brake");
    puts("");
    puts("For more keys, read the source code.");
    puts("");
}

int
main(int argc, char **argv)
{
    (void)argc;
    (void)argv;

    hello();
    gl_init();
    map_load_images();
    map_build();
    crosshair_build();
    gl_main_loop();

    return 0;
}
